indexing

	description:

		"Generates a reflection library"

	library: "Gobo Eiffel Tools Library"
	copyright: "Copyright (c) 2004, Eric Bezault and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date$"
	revision: "$Revision$"

class ERL_G_GENERATOR

inherit

	KL_SHARED_STREAMS
	ET_SHARED_CLASS_NAME_TESTER
	KL_SHARED_FILE_SYSTEM
	KL_IMPORTED_INTEGER_ROUTINES

creation

	make

feature {NONE} -- Initialization

	make (a_universe: like universe) is
			-- Create new reflection universe based on the gelint
			-- universe `a_universe'.
		require
			a_universe_not_void: a_universe /= Void
		do
			universe := a_universe
			create base_type_set.make
		ensure
			universe_set: universe = a_universe
		end

feature -- Status report

	has_fatal_error: BOOLEAN
			-- Has a fatal error occurred while generating?

	universe: ET_UNIVERSE
			-- Gelint universe the reflectable universe is based on

feature -- Status Report

	is_base_type_reflectable (a_base_type: ET_BASE_TYPE): BOOLEAN is
			-- Is base type `a_base_type' reflectable ?
		require
			a_base_type_not_void: a_base_type /= Void
			a_base_type_is_valid_context: a_base_type.is_valid_context
		local
			cs: DS_LINEAR_CURSOR [DS_PAIR [ET_BASE_TYPE, STRING]]
			other: ET_BASE_TYPE
		do
			from
				cs := base_type_set.new_cursor
				cs.start
			until
				cs.off or Result
			loop
				other := cs.item.first
				if a_base_type.same_base_type (other, other, a_base_type, universe) then
					Result := True
				end
				cs.forth
			end
			cs.go_after
		end

feature -- Access

	reflection_class_name_for_base_type (a_base_type: ET_BASE_TYPE): STRING is
			-- Name of reflection class for base type `a_base_type'
		require
			a_base_type_not_void: a_base_type /= Void
			a_base_type_reflectable: is_base_type_reflectable (a_base_type)
			a_base_type_is_valid_context: a_base_type.is_valid_context
		local
			cs: DS_LINEAR_CURSOR [DS_PAIR [ET_BASE_TYPE, STRING]]
			other: ET_BASE_TYPE
		do
			from
				cs := base_type_set.new_cursor
				cs.start
			until
				cs.off or Result /= Void
			loop
				other := cs.item.first
				if a_base_type.same_base_type (other, other, a_base_type, universe) then
					Result := cs.item.second
				end
				cs.forth
			end
			cs.go_after
		ensure
			name_not_void: Result /= Void
			name_not_empty: Result.count > 0
		end

feature -- Element change

	mark_base_type_reflectable (a_base_type: ET_BASE_TYPE) is
			-- Mark `a_base_type' reflectable.  Note that types
			-- `a_base_type' depends on will be marked reflectable as
			-- well (if they haven't been marked already).
		require
			a_base_type_not_void: a_base_type /= Void
			a_base_type_is_valid_context: a_base_type.is_valid_context
			a_base_type_not_yet_marked: not is_base_type_reflectable (a_base_type)
		local
			reflection_class_name: STRING
			base_class: ET_CLASS
			pair: DS_PAIR [ET_BASE_TYPE, STRING]
		do
    		base_class := a_base_type.direct_base_class (universe)
			reflection_class_name := "ERL_TYPE_IMP_"
			INTEGER_.append_decimal_integer (base_type_set.count + 1, reflection_class_name)
			reflection_class_name.append_character ('_')
			reflection_class_name.append_string (base_class.name.name)
			create pair.make (a_base_type, reflection_class_name)
			base_type_set.force_last (pair)
			mark_dependend_base_types_reflectable (a_base_type)
		ensure
			a_base_type_marked: is_base_type_reflectable (a_base_type)
		end

feature -- Generation

	generate_all (a_pathname: STRING) is
			-- Generate both the reflection universe class and the
			-- reflection type classes for all types marked reflectible.
			-- Generate all files into the directory `a_pathname'.
			-- Set `has_fatal_error' to `True' if generation cannot be
			-- completed.
		require
			a_pathname_not_void: a_pathname /= Void
		do
			generate_reflection_universe_class (a_pathname)
			generate_reflection_type_classes (a_pathname)
		end

	generate_reflection_universe_class (a_pathname: STRING) is
			-- Generate the reflection universe class.
			-- Generate the file into the directory `a_pathname'.
			-- Set `has_fatal_error' to `True' if generation cannot be
			-- completed.
		require
			a_pathname_not_void: a_pathname /= Void
		local
			reflection_universe_generator: ERL_G_UNIVERSE_GENERATOR
			file: KL_TEXT_OUTPUT_FILE
		do
			create reflection_universe_generator.make (Current)
			create file.make (file_system.pathname (a_pathname, "erl_universe_imp.e"))
			file.recursive_open_write
			if file.is_open_write then
				reflection_universe_generator.generate (file)
				file.close
				if reflection_universe_generator.has_fatal_error then
					has_fatal_error := True
				end
			else
				has_fatal_error := True
			end
		end

	generate_reflection_type_classes  (a_pathname: STRING) is
			-- Generate reflection type classes for all types makred reflectible.
			-- Generate all files into the directory `a_pathname'.
			-- Set `has_fatal_error' to `True' if generation cannot be
			-- completed.
		require
			a_pathname_not_void: a_pathname /= Void
		local
			cs: DS_LINEAR_CURSOR [DS_PAIR [ET_BASE_TYPE, STRING]]
			reflection_type_generator: ERL_G_TYPE_GENERATOR
			file: KL_TEXT_OUTPUT_FILE
			filename: STRING
		do
			create reflection_type_generator.make (Current)
			from
				cs := base_type_set.new_cursor
				cs.start
			until
				cs.off
			loop
				filename := cs.item.second.as_lower
				filename.append_string (".e")
				create file.make (file_system.pathname (a_pathname, filename))
				file.recursive_open_write
				if file.is_open_write then
					reflection_type_generator.generate (file, cs.item.first)
					file.close
					if reflection_type_generator.has_fatal_error then
						has_fatal_error := True
					end
				else
					has_fatal_error := True
				end
				cs.forth
			end
		end

feature {NONE} -- Implementation

	mark_dependend_base_types_reflectable (a_base_type: ET_BASE_TYPE) is
			-- Mark types `a_base_type' depends on reflectable (if they
			-- are not already reflectable).
		require
			a_base_type_not_void: a_base_type /= Void
			a_base_type_is_valid_context: a_base_type.is_valid_context
			a_base_type_marked: is_base_type_reflectable (a_base_type)
		local
			base_class: ET_CLASS
			feature_: ET_FEATURE
			creator: ET_CREATOR
			creator_index: INTEGER
			feature_index: INTEGER
		do
			base_class := a_base_type.context_base_class (universe)
			from
				feature_index := 1
			until
				feature_index > base_class.features.count
			loop
				feature_ := base_class.features.item (feature_index)
				if
					feature_.clients.has_class (universe.any_class) and
					not feature_.name.is_infix
				then
					mark_dependend_base_types_of_feature_reflectable (feature_, a_base_type)
				end
				feature_index := feature_index + 1
			end
			if base_class.creators /= Void then
				from
					creator_index := 1
				until
					creator_index > base_class.creators.count
				loop
					creator := base_class.creators.item (creator_index)
					if creator.clients.has_class (universe.any_class) then
						from
							feature_index := 1
						until
							feature_index > creator.count
						loop
							feature_ ?= base_class.named_feature (creator.feature_name (feature_index))
								check
									feature_not_void: feature_ /= Void
								end
							mark_dependend_base_types_of_feature_reflectable (feature_, a_base_type)
							feature_index := feature_index + 1
						end
					end
					creator_index := creator_index + 1
				end
			end
		end

	mark_dependend_base_types_of_feature_reflectable (a_feature: ET_FEATURE; a_base_type: ET_BASE_TYPE) is
			-- Mark base types which the feature `a_feature' when looked
			-- at from `a_base_type' is depending on as reflectible.
		require
			a_feature_not_void: a_feature /= Void
			a_base_type_not_void: a_base_type /= Void
			a_base_type_is_valid_context: a_base_type.is_valid_context
		local
			i: INTEGER
			type: ET_BASE_TYPE
		do
			if a_feature.type /= Void then
				type := a_feature.type.base_type (a_base_type, universe)
				if not is_base_type_reflectable (type) then
					mark_base_type_reflectable (type)
				end
			end
			if a_feature.arguments /= Void then
				from
					i := 1
				until
					i > a_feature.arguments.count
				loop
					type := a_feature.arguments.formal_argument (i).type.base_type (a_base_type, universe)
					if not is_base_type_reflectable (type) then
						mark_base_type_reflectable (type)
					end
					i := i + 1
				end
			end
		end

feature {ERL_G_UNIVERSE_GENERATOR} -- Implementation

	base_type_set: DS_LINKED_LIST [DS_PAIR [ET_BASE_TYPE, STRING]]
			-- Set of tuples containing the reflectable base types and
			-- their corresponding reflection class names

feature {NONE} -- Assertions helpers

	is_valid_pair (a_pair: DS_PAIR [ET_BASE_TYPE, STRING]): BOOLEAN is
			-- Is `a_pair's valid?
		do
			Result := a_pair /= Void and then
					a_pair.first /= Void and then
					a_pair.second /= Void and then
					a_pair.first.is_valid_context
		ensure
			definition: Result = (a_pair /= Void and then
										 a_pair.first /= Void and then
										 a_pair.second /= Void and then
										 a_pair.first.is_valid_context)
		end

	base_types_are_valid_contexts: BOOLEAN is
			-- Are all types in `base_type_set' valid contexts?
		local
			cs: DS_LINEAR_CURSOR [DS_PAIR [ET_BASE_TYPE, STRING]]
		do
			Result := True
			from
				cs := base_type_set.new_cursor
				cs.start
			until
				cs.off or not Result
			loop
				if not is_valid_pair (cs.item) then
					Result := False
				end
			end
			cs.go_after
		end

invariant

	universe_not_void: universe /= Void
	base_type_set_not_void: base_type_set /= Void
	base_types_are_valid_context: base_types_are_valid_contexts

end
